unit DrawObject;
(*
 * Concave Polygon Scan Conversion
 * by Paul Heckbert
 * from "Graphics Gems", Academic Press, 1990
 *)

interface

uses
  Classes, Windows, Graphics
  ;

type

  TEdge = record                        // a polygon edge
    x: integer;                         // x coordinate of edge's intersection with current scanline
    Inc: integer;
    Tst: integer;
    Acc: integer;
    Step: integer;
    i: integer;                         // edge number: edge i goes from pt[i] to pt[i+1]
  end;
  PEdge = ^TEdge;
  TEdgeArray = array of TEdge;

  TPointArray = array of TPoint;

  TIntegerArray = array[0..0] of integer;
  PIntegerArray = ^TIntegerArray;

  TWordArray = array[0..0] of integer;
  PWordArray = ^TWordArray;

  TFastDraw = class(TObject)
  private
    fBitmap: TBitmap;
    fClipRect: TRect;
    fLinesPtr: array of Pointer;
    fPitch: integer;
    fnAct: integer;
    fActiveEdge: array of TEdge;
    fnPts: integer;
    fIs16: boolean;
  protected
    procedure RecreateScanLines; virtual; abstract;
    procedure CDelete(i: integer);
    procedure CInsert(pt: array of TPoint; i, y: integer);
  public
    constructor Create(aBitmap: TBitmap); virtual;
    destructor Destroy; override;

    procedure LineP(P1, P2: TPoint); virtual; abstract;
    procedure LineC(X1, Y1, X2, Y2: integer); virtual; abstract;
    procedure LineTo(X, Y: Integer);
    procedure MoveTo(X, Y: Integer);
    procedure Polygon(const Points: array of TPoint); virtual; abstract;
    procedure Polyline(const Points: array of TPoint); virtual; abstract;
    procedure PolylineClosed(const Points: array of TPoint); virtual; abstract;
    procedure Rectangle(X1, Y1, X2, Y2: integer);
  end;

  TFastDraw32 = class(TFastDraw)
  private
    fPenColor32: longword;
    fBrushColor32: longword;
    FillSub: procedure(X: Plongword; Count: Integer; Value: Longword);
    FillOneSub: procedure(X: Plongword; Value: Longword);
  protected
    procedure RecreateScanLines; override;
  public
    constructor Create(aBitmap: TBitmap); override;
    destructor Destroy; override;

    procedure LineP(P1, P2: TPoint); override;
    procedure LineC(X1, Y1, X2, Y2: integer); override;
    procedure Polygon(const Points: array of TPoint); override;
    procedure Polyline(const Points: array of TPoint); override;
    procedure PolylineClosed(const Points: array of TPoint); override;
  end;

  TFastDraw24 = class(TFastDraw)
  private
    fPenColor24: longword;
    fBrushColor24: longword;
    FillSub: procedure(X: PRGBTriple; Count: Integer; Value: Longword);
    FillOneSub: procedure(X: PRGBTriple; Value: Longword);
  protected
    procedure RecreateScanLines; override;
  public
    constructor Create(aBitmap: TBitmap); override;
    destructor Destroy; override;

    procedure LineP(P1, P2: TPoint); override;
    procedure LineC(X1, Y1, X2, Y2: integer); override;
    procedure Polygon(const Points: array of TPoint); override;
    procedure Polyline(const Points: array of TPoint); override;
    procedure PolylineClosed(const Points: array of TPoint); override;
  end;

  TFastDraw15 = class(TFastDraw)
  private
    fPenColor15: word;
    fBrushColor15: word;
    FillSub: procedure(X: Pword; Count: Integer; Value: word);
    FillOneSub: procedure(X: Pword; Value: word);
  protected
    procedure RecreateScanLines; override;
  public
    constructor Create(aBitmap: TBitmap); override;
    destructor Destroy; override;

    procedure LineP(P1, P2: TPoint); override;
    procedure LineC(X1, Y1, X2, Y2: integer); override;
    procedure Polygon(const Points: array of TPoint); override;
    procedure Polyline(const Points: array of TPoint); override;
    procedure PolylineClosed(const Points: array of TPoint); override;
  end;

  TDrawObject = class(TObject)
  private
    fFastDraw: TFastDraw;
    fCanvas: TCanvas;
    procedure SetupCanvas(aCanvas: TCanvas);
    procedure LineC(X1, Y1, X2, Y2: integer);
  public
    CopyRect: procedure(const Dest: TRect; Canvas: TCanvas; const Source: TRect) of object;
    Draw: procedure(X, Y: Integer; Graphic: TGraphic) of object;
    Ellipse: procedure(X1, Y1, X2, Y2: Integer) of object;
    FillRect: procedure(const Rect: TRect) of object;
    FrameRect: procedure(const Rect: TRect) of object;
    Line: procedure(X1, Y1, X2, Y2: Integer) of object;
    LineTo: procedure(X, Y: Integer) of object;
    MoveTo: procedure(X, Y: Integer) of object;
    Polygon: procedure(const Points: array of TPoint) of object;
    Polyline: procedure(const Points: array of TPoint) of object;
    Rectangle: procedure(X1, Y1, X2, Y2: integer) of object;
    Refresh: procedure of object;
    StretchDraw: procedure(const Rect: TRect; Graphic: TGraphic) of object;
    TextHeight: function(const Text: string): Integer of object;
    TextOut: procedure(X, Y: Integer; const Text: string) of object;
    TextRect: procedure(Rect: TRect; X, Y: Integer; const Text: string) of object;
    TextWidth: function(const Text: string): Integer of object;

    constructor Create(aCanvas: TCanvas); overload;
    constructor Create(aBitmap: TBitmap); overload;
    destructor Destroy; override;

    procedure BeginUpdate;
    procedure EndUpdate;

    property Canvas: TCanvas read fCanvas;
  end;

function Color32(c: TColor): integer;

implementation

uses
  Math
  ;

// ---------------------------------------------------------------------

function Sgn(x: integer): integer;
begin
  if x < 0 then
    Result := -1
  else if x = 0 then
    Result := 0
  else
    Result := 1
end;

// ---------------------------------------------------------------------

{ TFastDraw }

constructor TFastDraw.Create(aBitmap: TBitmap);
begin
  fBitmap := aBitmap;
  RecreateScanLines;
end;

destructor TFastDraw.Destroy;
begin
  fBitmap := nil;
  inherited;
end;

// ---------------------------------------------------------------------

procedure TFastDraw.CDelete(i: integer); // remove edge i from ActiveEdge list
var
  j                 : integer;
begin
  j := 0;
  while (j < fnAct) and (fActiveEdge[j].i <> i) do
    Inc(j);

  if (j >= fnAct) then Exit;            // edge not in ActiveEdge list; happens at win->y0
  Dec(fnAct);
  Move(fActiveEdge[j + 1], fActiveEdge[j], (fnAct - j) * SizeOf(fActiveEdge[0]));
end;

procedure TFastDraw.CInsert(pt: array of TPoint; i, y: integer); // append edge i to end of ActiveEdge list */
var
  j                 : integer;
  dx                : integer;
  dy                : integer;
  p, q              : PPoint;
begin
  if (i < (fnPts - 1)) then
    j := i + 1
  else
    j := 0;

  if (pt[i].y < pt[j].y) then begin
    p := @pt[i];
    q := @pt[j];
  end
  else begin
    p := @pt[j];
    q := @pt[i];
  end;

  // initialize x position at intersection of edge with scanline y */
  dx := q.x - p.x;
  dy := q.y - p.y;

  fActiveEdge[fnAct].Acc := 0;
  if dx < 0 then begin
    fActiveEdge[fnAct].Inc := -dx;
    fActiveEdge[fnAct].Step := -1;
  end
  else begin
    fActiveEdge[fnAct].Inc := dx;
    fActiveEdge[fnAct].Step := 1;
  end;
  if dy < 0 then
    fActiveEdge[fnAct].Tst := -dy
  else
    fActiveEdge[fnAct].Tst := dy;
  fActiveEdge[fnAct].x := Floor(dx / dy * (y - p.y)) + p.x;
  fActiveEdge[fnAct].i := i;
  Inc(fnAct);
end;

procedure TFastDraw.LineTo(X, Y: integer);
begin
  LineP(fBitmap.Canvas.PenPos, Point(X, Y));
end;

procedure TFastDraw.MoveTo(X, Y: integer);
begin
  fBitmap.Canvas.PenPos := Point(X, Y);
end;

procedure TFastDraw.Rectangle(X1, Y1, X2, Y2: integer);
begin
  Polygon([Point(X1, Y1), Point(X1, Y2), Point(X2, Y2), Point(X2, Y1)]);
end;

(*
procedure TFastDraw.FillRect(const R: TRect);
var pc: TColor;
begin
  pc := fBitmap.Canvas.Pen.Color;
  fBitmap.Canvas.Pen.Color := fBitmap.Canvas.Brush.Color;
  Polygon([Point(R.Left, R.Top), Point(R.Left, R.Bottom),
           Point(R.Right, R.Bottom), Point(R.Right, R.Top)]);
end;
*)

{ TFastDraw32 }

procedure FillOne32AlfaBlend(X: Plongword; Value: Longword);
begin
  X^ := ((Value and $00FEFEFE) + (X^ and $00FEFEFE)) shr 1;
end;

procedure Fill32AlfaBlend(X: Plongword; Count: Integer; Value: Longword);
begin
  while Count > 0 do begin
    X^ := ((Value and $00FEFEFE) + (X^ and $00FEFEFE)) shr 1;
    Inc(X);
    Dec(Count);
  end
end;

procedure FillOne32(X: Plongword; Value: Longword);
begin
  X^ := Value;
end;

procedure Fill32(X: Plongword; Count: Integer; Value: Longword);
asm
// EAX = X
// EDX = Count
// ECX = Value
        PUSH    EDI

        MOV     EDI,EAX  // Point EDI to destination
        MOV     EAX,ECX
        MOV     ECX,EDX
        TEST    ECX,ECX
        JS      @@exit

        REP     STOSD    // Fill count dwords
@@exit:
        POP     EDI
end;

function Color32(c: TColor): integer;
var
  cl                : longword;
begin
  cl := ColorToRGB(c);
  Result := ((cl and $FF0000) shr 16) +
    ((cl and $00FF00)) +
    ((cl and $0000FF) shl 16);
end;

//-------------------------------------------------------------------------

procedure TFastDraw32.RecreateScanLines;
var
  i                 : integer;
begin
  SetLength(fLinesPtr, fBitmap.Height);
  for i := 0 to Pred(fBitmap.Height) do
    fLinesPtr[i] := fBitmap.Scanline[i];
  if Length(fLinesPtr) >= 2 then
    fPitch := (integer(fLinesPtr[1]) - integer(fLinesPtr[0])) div SizeOf(longword)
  else
    fPitch := 0;
  fClipRect := Rect(0, 0, fBitmap.Width, fBitmap.Height);
  Inc(fClipRect.Top);
  Inc(fClipRect.Left);
  Dec(fClipRect.Right);
  Dec(fClipRect.Bottom, 2);
end;

procedure TFastDraw32.LineP(p1, p2: TPoint);
var
  d, x, y, ax, ay, sx, sy, dx, dy: integer;
  pptr              : PLongWord;
  pyInc             : integer;
  Code1, Code2      : integer;

begin

  fPenColor32 := Color32(fBitmap.Canvas.Pen.Color);

  dx := p2.x - p1.x;
  dy := p2.y - p1.y;

  //  if ClipLine(p1, p2, fClipRect, dx, dy) then begin
  Code1 := 0;
  Code2 := 0;

  if (p1.x < fClipRect.Left) then begin
    if dx <> 0 then begin
      p1.y := Floor(p1.y + (dy / dx) * (fClipRect.Left - p1.x));
      p1.x := fClipRect.Left;
    end;
    Code1 := 1;
  end
  else if (p1.x > fClipRect.Right) then begin
    if dx <> 0 then begin
      p1.y := Floor(p1.y + (dy / dx) * (fClipRect.Right - p1.x));
      p1.x := fClipRect.Right;
    end;
    Code1 := 2;
  end;

  if (p1.y < fClipRect.Top) then begin
    if dy <> 0 then begin
      p1.x := Floor(p1.x + (dx / dy) * (fClipRect.Top - p1.y));
      p1.y := fClipRect.Top;
    end;
    Code1 := Code1 or 4;
  end
  else if (p1.y > fClipRect.Bottom) then begin
    if dy <> 0 then begin
      p1.x := Floor(p1.x + (dx / dy) * (fClipRect.Bottom - p1.y));
      p1.y := fClipRect.Bottom;
    end;
    Code1 := Code1 or 8;
  end;

  if (p2.x < fClipRect.Left) then begin
    if dx <> 0 then begin
      p2.y := Floor(p2.y + (dy / dx) * (fClipRect.Left - p2.x));
      p2.x := fClipRect.Left;
    end;
    Code2 := 1;
  end
  else if (p2.x > fClipRect.Right) then begin
    if dx <> 0 then begin
      p2.y := Floor(p2.y + (dy / dx) * (fClipRect.Right - p2.x));
      p2.x := fClipRect.Right;
    end;
    Code2 := 2;
  end;

  if (p2.y < fClipRect.Top) then begin
    if dy <> 0 then begin
      p2.x := Floor(p2.x + (dx / dy) * (fClipRect.Top - p2.y));
      p2.y := fClipRect.Top;
    end;
    Code2 := Code2 or 4;
  end
  else if (p2.y > fClipRect.Bottom) then begin
    if dy <> 0 then begin
      p2.x := Floor(p2.x + (dx / dy) * (fClipRect.Bottom - p2.y));
      p2.y := fClipRect.Bottom;
    end;
    Code2 := Code2 or 8;
  end;

  if ((Code1 and Code2) = 0) and
    (p1.x >= fClipRect.Left) and
    (p1.x <= fClipRect.Right) then begin

    // ---------------------------------------------------
    ax := Abs(dx) shl 1;
    ay := Abs(dy) shl 1;

    sx := Sgn(dx);
    sy := Sgn(dy);

    pyInc := sy * fPitch;

    x := p1.x;
    y := p1.y;
    pptr := fLinesPtr[y];
    Inc(pptr, x);
    if (ax > ay) then begin             (* x dominant *)
      d := ay - (ax shr 1);
      if dx < 0 then begin
        while (x >= p2.x) do begin
          //          pptr^ := fPenColor32;
          FillOneSub(pptr, fPenColor32);
          //          if (x = p2.x) then
          //            Exit;
          if (d >= 0) then begin
            //            Inc(y, sy);
            Inc(pptr, pyInc);
            Dec(d, ax);
          end;
          Dec(x);
          Dec(pptr);
          Inc(d, ay);
        end
      end
      else begin
        while (x <= p2.x) do begin
          //          pptr^ := fPenColor32;
          FillOneSub(pptr, fPenColor32);
          //          if (x = p2.x) then
          //            Exit;
          if (d >= 0) then begin
            //            Inc(y, sy);
            Inc(pptr, pyInc);
            Dec(d, ax);
          end;
          Inc(x);
          Inc(pptr);
          Inc(d, ay);
        end
      end
    end
    else begin                          (* y dominant *)
      d := ax - (ay shr 1);
      if dy < 0 then begin
        while (y >= p2.y) do begin
          //          pptr^ := fPenColor32;
          FillOneSub(pptr, fPenColor32);
          //          if (y = p2.y) then
          //            Exit;
          if (d >= 0) then begin
            //            Inc(x, sx);
            Inc(pptr, sx);
            Dec(d, ay);
          end;
          Dec(y);
          Inc(pptr, pyInc);
          Inc(d, ax);
        end
      end
      else begin
        while (y <= p2.y) do begin
          //          pptr^ := fPenColor32;
          FillOneSub(pptr, fPenColor32);
          //          if (y = p2.y) then
          //            Exit;
          if (d >= 0) then begin
            //            Inc(x, sx);
            Inc(pptr, sx);
            Dec(d, ay);
          end;
          Inc(y);
          Inc(pptr, pyInc);
          Inc(d, ax);
        end
      end
    end
  end
end;

procedure TFastDraw32.LineC(x1, y1, x2, y2: integer);
var
  d, x, y, ax, ay, sx, sy, dx, dy: integer;
  pptr              : Plongword;
  pyInc             : integer;
  Code1, Code2      : integer;

begin

  fPenColor32 := Color32(fBitmap.Canvas.Pen.Color);

  dx := x2 - x1;
  dy := y2 - y1;

  //  if ClipLine(p1, p2, fClipRect, dx, dy) then begin
  Code1 := 0;
  Code2 := 0;

  if (x1 < fClipRect.Left) then begin
    if dx <> 0 then begin
      y1 := Floor(y1 + (dy / dx) * (fClipRect.Left - x1));
      x1 := fClipRect.Left;
    end;
    Code1 := 1;
  end
  else if (x1 > fClipRect.Right) then begin
    if dx <> 0 then begin
      y1 := Floor(y1 + (dy / dx) * (fClipRect.Right - x1));
      x1 := fClipRect.Right;
    end;
    Code1 := 2;
  end;

  if (y1 < fClipRect.Top) then begin
    if dy <> 0 then begin
      x1 := Floor(x1 + (dx / dy) * (fClipRect.Top - y1));
      y1 := fClipRect.Top;
    end;
    Code1 := Code1 or 4;
  end
  else if (y1 > fClipRect.Bottom) then begin
    if dy <> 0 then begin
      x1 := Floor(x1 + (dx / dy) * (fClipRect.Bottom - y1));
      y1 := fClipRect.Bottom;
    end;
    Code1 := Code1 or 8;
  end;

  if (x2 < fClipRect.Left) then begin
    if dx <> 0 then begin
      y2 := Floor(y2 + (dy / dx) * (fClipRect.Left - x2));
      x2 := fClipRect.Left;
    end;
    Code2 := 1;
  end
  else if (x2 > fClipRect.Right) then begin
    if dx <> 0 then begin
      y2 := Floor(y2 + (dy / dx) * (fClipRect.Right - x2));
      x2 := fClipRect.Right;
    end;
    Code2 := 2;
  end;

  if (y2 < fClipRect.Top) then begin
    if dy <> 0 then begin
      x2 := Floor(x2 + (dx / dy) * (fClipRect.Top - y2));
      y2 := fClipRect.Top;
    end;
    Code2 := Code2 or 4;
  end
  else if (y2 > fClipRect.Bottom) then begin
    if dy <> 0 then begin
      x2 := Floor(x2 + (dx / dy) * (fClipRect.Bottom - y2));
      y2 := fClipRect.Bottom;
    end;
    Code2 := Code2 or 8;
  end;

  if ((Code1 and Code2) = 0) and
    (x1 >= fClipRect.Left) and
    (x1 <= fClipRect.Right) then begin

    // ---------------------------------------------------
    ax := Abs(dx) shl 1;
    ay := Abs(dy) shl 1;

    sx := Sgn(dx);
    sy := Sgn(dy);

    pyInc := sy * fPitch;

    x := x1;
    y := y1;
    pptr := fLinesPtr[y];
    Inc(pptr, x);
    if (ax > ay) then begin             (* x dominant *)
      d := ay - (ax shr 1);
      if dx < 0 then begin
        while (x >= x2) do begin
          //        pptr^ := fPenColor32;
          FillOneSub(pptr, fPenColor32);
          //          if (x = x2) then
          //            Exit;
          if (d >= 0) then begin
            //            Inc(y, sy);
            Inc(pptr, pyInc);
            Dec(d, ax);
          end;
          Dec(x);
          Dec(pptr);
          Inc(d, ay);
        end
      end
      else begin
        while (x <= x2) do begin
          //        pptr^ := fPenColor32;
          FillOneSub(pptr, fPenColor32);
          //          if (x = x2) then
          //            Exit;
          if (d >= 0) then begin
            //            Inc(y, sy);
            Inc(pptr, pyInc);
            Dec(d, ax);
          end;
          Inc(x);
          Inc(pptr);
          Inc(d, ay);
        end
      end
    end
    else begin                          (* y dominant *)
      d := ax - (ay shr 1);
      if dy < 0 then begin
        while (y >= y2) do begin
          //        pptr^ := fPenColor32;
          FillOneSub(pptr, fPenColor32);
          //          if (y = y2) then
          //            Exit;
          if (d >= 0) then begin
            //            Inc(x, sx);
            Inc(pptr, sx);
            Dec(d, ay);
          end;
          Dec(y);
          Inc(pptr, pyInc);
          Inc(d, ax);
        end
      end
      else begin
        while (y <= y2) do begin
          //        pptr^ := fPenColor32;
          FillOneSub(pptr, fPenColor32);
          //          if (y = y2) then
          //            Exit;
          if (d >= 0) then begin
            //            Inc(x, sx);
            Inc(pptr, sx);
            Dec(d, ay);
          end;
          Inc(y);
          Inc(pptr, pyInc);
          Inc(d, ax);
        end
      end
    end
  end;
  fBitmap.Canvas.PenPos := Point(X2, Y2);
end;

procedure TFastDraw32.PolyLine(const Points: array of TPoint);
var
  i                 : integer;
begin
  fPenColor32 := Color32(fBitmap.Canvas.Pen.Color);
  fBrushColor32 := Color32(fBitmap.Canvas.Brush.Color);

  for i := 0 to Pred(High(Points)) do begin
    LineP(Points[i], Points[i + 1]);
  end;
  fBitmap.Canvas.PenPos := Points[High(Points)];
end;

procedure TFastDraw32.PolyLineClosed(const Points: array of TPoint);
var
  i                 : integer;
begin
  fPenColor32 := Color32(fBitmap.Canvas.Pen.Color);
  fBrushColor32 := Color32(fBitmap.Canvas.Brush.Color);

  for i := 0 to Pred(High(Points)) do begin
    LineP(Points[i], Points[i + 1]);
  end;
  LineP(Points[High(Points)], Points[0]);
  fBitmap.Canvas.PenPos := Points[0];
end;

procedure TFastDraw32.Polygon(const Points: array of TPoint);
var
  a, b, t           : integer;
  k, y0, y1, y, i, j, xl, xr: integer;
  Ind               : array of integer; // list of vertex indices, sorted by pt[ind[j]].y
  e                 : TEdge;
  ep                : PEdge;
begin

  fPenColor32 := Color32(fBitmap.Canvas.Pen.Color);
  fBrushColor32 := Color32(fBitmap.Canvas.Brush.Color);

  if fBitmap.Canvas.Brush.Style = bsDiagCross then begin
    FillSub := Fill32AlfaBlend;
    FillOneSub := FillOne32AlfaBlend;
  end
  else begin
    FillSub := Fill32;
    FillOneSub := FillOne32;
  end;

  fnPts := Succ(High(Points));

  if (fnPts > 2) and (fBitmap.Canvas.Brush.Style <> bsClear) then begin

    SetLength(Ind, fnPts);
    SetLength(fActiveEdge, fnPts);

    // create y-sorted array of indices ind[k] into vertex list
    for k := 0 to Pred(fnPts) do
      Ind[k] := k;

    for a := 0 to Pred(Pred(fnPts)) do begin
      for b := Succ(a) to Pred(fnPts) do begin
        if Points[Ind[b]].y <= Points[Ind[a]].y then begin
          t := Ind[b];
          Ind[b] := Ind[a];
          Ind[a] := t;
        end
      end
    end;

    fnAct := 0;                         // start with empty ActiveEdge list
    y0 := Max(fClipRect.Top, Points[ind[0]].y); // ymin of polygon
    y1 := Min(fClipRect.Bottom, Points[ind[fnPts - 1]].y); // ymax of polygon

    y := y0;
    k := 0;                             // ind[k] is next vertex to process
    while y <= y1 do begin              // step through scanlines
      // check vertices between previous scanline and current one, if any
      while (k < fnPts) and (Points[ind[k]].y <= y) do begin
        i := Ind[k];
        // insert or delete edges before and after vertex i (i-1 to i,
        // and i to i+1) from ActiveEdge list if they cross scanline y
        if (i > 0) then
          j := i - 1
        else
          j := fnPts - 1;               // vertex previous to i

        if (Points[j].y < y) then       // old edge, remove from ActiveEdge list
          CDelete(j)
        else if (Points[j].y > y) then  // new edge, add to ActiveEdge list
          CInsert(Points, j, y);

        if (i < (fnPts - 1)) then
          j := i + 1
        else
          j := 0;                       // vertex next after i

        if (Points[j].y < y) then       // old edge, remove from ActiveEdge list
          CDelete(i)
        else if (Points[j].y > y) then  // new edge, add to ActiveEdge list
          CInsert(Points, i, y);

        Inc(k);
      end;

      // sort ActiveEdge edge list by ActiveEdge[j].x
      for a := 0 to Pred(Pred(fnAct)) do begin
        for b := Succ(a) to Pred(fnAct) do begin
          if fActiveEdge[b].x <= fActiveEdge[a].x then begin
            e := fActiveEdge[b];
            fActiveEdge[b] := fActiveEdge[a];
            fActiveEdge[a] := e;
          end
        end
      end;
      // draw horizontal segments for scanline y
      j := 0;
      while (j < fnAct) do begin        // draw horizontal segments

        // span 'tween j & j+1 is inside, span tween j+1 & j+2 is outside
        xl := fActiveEdge[j].x;         // left end of span
        if (xl < fClipRect.Left) then xl := fClipRect.Left;
        xr := fActiveEdge[j + 1].x;     // right end of span
        if (xr > fClipRect.Right) then xr := fClipRect.Right;
        if (xl <= xr) then begin
          // draw pixels in span
          (**)
          FillSub(@PIntegerArray(fLinesPtr[y])[xl], xr - xl + 1, fBrushColor32);
          (**)
          (*
                  pptr := fLinesPtr[y];
                  Inc(pptr, xl);
                  while xl <= xr do begin
                    pptr^ := cb;
                    Inc(pptr);
                    Inc(xl);
                  end;
          *)
        end;

        ep := @fActiveEdge[j];
        ep.Acc := ep.Acc + Abs(ep.Inc);
        while ep.Acc >= ep.Tst do begin
          Inc(ep.x, ep.Step);
          ep.Acc := ep.Acc - ep.Tst;
        end;
        Inc(j);

        ep := @fActiveEdge[j];
        Inc(ep.Acc, ep.Inc);
        while ep.Acc >= ep.Tst do begin
          Inc(ep.x, ep.Step);
          ep.Acc := ep.Acc - ep.Tst;
        end;
        Inc(j);

      end;
      Inc(y);
    end
  end
  else begin
    fBrushColor32 := $FFFFFFFF;
  end;
  PolyLineClosed(Points);
  fBitmap.Canvas.PenPos := Points[High(Points)];
end;

constructor TFastDraw32.Create(aBitmap: TBitmap);
begin
  inherited Create(aBitmap);
end;

destructor TFastDraw32.Destroy;
begin
  inherited Destroy;
end;

{ TFastDraw24 }

type
  TTripleInt = record
    case boolean of
      true: (RGB: TRGBTriple);
      false: (i: longint);
  end;
  PTripleInt = ^TTripleInt;

  TRGBTripleArray = array[0..0] of TRGBTriple;
  PRGBTripleArray = ^TRGBTripleArray;

procedure FillOne24AlfaBlend(X: PRGBTriple; Value: Longword);
var
  Res: longint;
begin
  Res := ((Value and $FEFEFE) + (PTripleInt(X)^.i and $FEFEFE)) shr 1;
  X^ := PTripleInt(@Res).RGB;
end;

procedure Fill24AlfaBlend(X: PRGBTriple; Count: Integer; Value: Longword);
var
  Res: longint;
begin
  while Count > 0 do begin
    Res := ((Value and $FEFEFE) + (PTripleInt(X)^.i and $FEFEFE)) shr 1;
    X^ := PTripleInt(@Res).RGB;
    Inc(X);
    Dec(Count);
  end
end;

procedure FillOne24(X: PRGBTriple; Value: Longword);
begin
  X^ := PTripleInt(@Value).RGB;
end;

procedure Fill24(X: PRGBTriple; Count: Integer; Value: Longword);
begin
  while Count > 0 do begin
    X^ := PTripleInt(@Value).RGB;
    Inc(X);
    Dec(Count);
  end
end;

function Color24(c: TColor): integer;
var
  cl                : longword;
begin
  cl := ColorToRGB(c);
  Result := ((cl and $FF0000) shr 16) +
    ((cl and $00FF00)) +
    ((cl and $0000FF) shl 16);
end;

//-------------------------------------------------------------------------

procedure TFastDraw24.RecreateScanLines;
var
  i                 : integer;
begin
  SetLength(fLinesPtr, fBitmap.Height);
  for i := 0 to Pred(fBitmap.Height) do
    fLinesPtr[i] := fBitmap.Scanline[i];
  if Length(fLinesPtr) >= 2 then
    fPitch := (integer(fLinesPtr[1]) - integer(fLinesPtr[0]))
  else
    fPitch := 0;
  fClipRect := Rect(0, 0, fBitmap.Width, fBitmap.Height);
  Inc(fClipRect.Top);
  Inc(fClipRect.Left);
  Dec(fClipRect.Right);
  Dec(fClipRect.Bottom, 2);
end;

procedure TFastDraw24.LineP(p1, p2: TPoint);
var
  d, x, y, ax, ay, sx, sy, dx, dy: integer;
  pptr              : PRGBTriple;
  pyInc             : integer;
  Code1, Code2      : integer;

begin
  fPenColor24 := Color24(fBitmap.Canvas.Pen.Color);

  dx := p2.x - p1.x;
  dy := p2.y - p1.y;

  //  if ClipLine(p1, p2, fClipRect, dx, dy) then begin
  Code1 := 0;
  Code2 := 0;

  if (p1.x < fClipRect.Left) then begin
    if dx <> 0 then begin
      p1.y := Floor(p1.y + (dy / dx) * (fClipRect.Left - p1.x));
      p1.x := fClipRect.Left;
    end;
    Code1 := 1;
  end
  else if (p1.x > fClipRect.Right) then begin
    if dx <> 0 then begin
      p1.y := Floor(p1.y + (dy / dx) * (fClipRect.Right - p1.x));
      p1.x := fClipRect.Right;
    end;
    Code1 := 2;
  end;

  if (p1.y < fClipRect.Top) then begin
    if dy <> 0 then begin
      p1.x := Floor(p1.x + (dx / dy) * (fClipRect.Top - p1.y));
      p1.y := fClipRect.Top;
    end;
    Code1 := Code1 or 4;
  end
  else if (p1.y > fClipRect.Bottom) then begin
    if dy <> 0 then begin
      p1.x := Floor(p1.x + (dx / dy) * (fClipRect.Bottom - p1.y));
      p1.y := fClipRect.Bottom;
    end;
    Code1 := Code1 or 8;
  end;

  if (p2.x < fClipRect.Left) then begin
    if dx <> 0 then begin
      p2.y := Floor(p2.y + (dy / dx) * (fClipRect.Left - p2.x));
      p2.x := fClipRect.Left;
    end;
    Code2 := 1;
  end
  else if (p2.x > fClipRect.Right) then begin
    if dx <> 0 then begin
      p2.y := Floor(p2.y + (dy / dx) * (fClipRect.Right - p2.x));
      p2.x := fClipRect.Right;
    end;
    Code2 := 2;
  end;

  if (p2.y < fClipRect.Top) then begin
    if dy <> 0 then begin
      p2.x := Floor(p2.x + (dx / dy) * (fClipRect.Top - p2.y));
      p2.y := fClipRect.Top;
    end;
    Code2 := Code2 or 4;
  end
  else if (p2.y > fClipRect.Bottom) then begin
    if dy <> 0 then begin
      p2.x := Floor(p2.x + (dx / dy) * (fClipRect.Bottom - p2.y));
      p2.y := fClipRect.Bottom;
    end;
    Code2 := Code2 or 8;
  end;

  if ((Code1 and Code2) = 0) and
    (p1.x >= fClipRect.Left) and
    (p1.x <= fClipRect.Right) then begin

    // ---------------------------------------------------
    ax := Abs(dx) shl 1;
    ay := Abs(dy) shl 1;

    sx := Sgn(dx);
    sy := Sgn(dy);

    pyInc := sy * fPitch;

    x := p1.x;
    y := p1.y;
    pptr := fLinesPtr[y];
    Inc(pptr, x);
    if (ax > ay) then begin             (* x dominant *)
      d := ay - (ax shr 1);
      if dx < 0 then begin
        while (x >= p2.x) do begin
          FillOneSub(pptr, fPenColor24);
          if (d >= 0) then begin
            Inc(PByte(pptr), pyInc);
            Dec(d, ax);
          end;
          Dec(x);
          Dec(pptr);
          Inc(d, ay);
        end
      end
      else begin
        while (x <= p2.x) do begin
          FillOneSub(pptr, fPenColor24);
          if (d >= 0) then begin
            Inc(PByte(pptr), pyInc);
            Dec(d, ax);
          end;
          Inc(x);
          Inc(pptr);
          Inc(d, ay);
        end
      end
    end
    else begin                          (* y dominant *)
      d := ax - (ay shr 1);
      if dy < 0 then begin
        while (y >= p2.y) do begin
          FillOneSub(pptr, fPenColor24);
          if (d >= 0) then begin
            Inc(pptr, sx);
            Dec(d, ay);
          end;
          Dec(y);
          Inc(PByte(pptr), pyInc);
          Inc(d, ax);
        end
      end
      else begin
        while (y <= p2.y) do begin
          FillOneSub(pptr, fPenColor24);
          if (d >= 0) then begin
            Inc(pptr, sx);
            Dec(d, ay);
          end;
          Inc(y);
          Inc(PByte(pptr), pyInc);
          Inc(d, ax);
        end
      end
    end
  end
end;

procedure TFastDraw24.LineC(x1, y1, x2, y2: integer);
var
  d, x, y, ax, ay, sx, sy, dx, dy: integer;
  pptr              : PRGBTriple;
  pyInc             : integer;
  Code1, Code2      : integer;

begin
  fPenColor24 := Color24(fBitmap.Canvas.Pen.Color);

  dx := x2 - x1;
  dy := y2 - y1;

  //  if ClipLine(p1, p2, fClipRect, dx, dy) then begin
  Code1 := 0;
  Code2 := 0;

  if (x1 < fClipRect.Left) then begin
    if dx <> 0 then begin
      y1 := Floor(y1 + (dy / dx) * (fClipRect.Left - x1));
      x1 := fClipRect.Left;
    end;
    Code1 := 1;
  end
  else if (x1 > fClipRect.Right) then begin
    if dx <> 0 then begin
      y1 := Floor(y1 + (dy / dx) * (fClipRect.Right - x1));
      x1 := fClipRect.Right;
    end;
    Code1 := 2;
  end;

  if (y1 < fClipRect.Top) then begin
    if dy <> 0 then begin
      x1 := Floor(x1 + (dx / dy) * (fClipRect.Top - y1));
      y1 := fClipRect.Top;
    end;
    Code1 := Code1 or 4;
  end
  else if (y1 > fClipRect.Bottom) then begin
    if dy <> 0 then begin
      x1 := Floor(x1 + (dx / dy) * (fClipRect.Bottom - y1));
      y1 := fClipRect.Bottom;
    end;
    Code1 := Code1 or 8;
  end;

  if (x2 < fClipRect.Left) then begin
    if dx <> 0 then begin
      y2 := Floor(y2 + (dy / dx) * (fClipRect.Left - x2));
      x2 := fClipRect.Left;
    end;
    Code2 := 1;
  end
  else if (x2 > fClipRect.Right) then begin
    if dx <> 0 then begin
      y2 := Floor(y2 + (dy / dx) * (fClipRect.Right - x2));
      x2 := fClipRect.Right;
    end;
    Code2 := 2;
  end;

  if (y2 < fClipRect.Top) then begin
    if dy <> 0 then begin
      x2 := Floor(x2 + (dx / dy) * (fClipRect.Top - y2));
      y2 := fClipRect.Top;
    end;
    Code2 := Code2 or 4;
  end
  else if (y2 > fClipRect.Bottom) then begin
    if dy <> 0 then begin
      x2 := Floor(x2 + (dx / dy) * (fClipRect.Bottom - y2));
      y2 := fClipRect.Bottom;
    end;
    Code2 := Code2 or 8;
  end;

  if ((Code1 and Code2) = 0) and
    (x1 >= fClipRect.Left) and
    (x1 <= fClipRect.Right) then begin

    // ---------------------------------------------------
    ax := Abs(dx) shl 1;
    ay := Abs(dy) shl 1;

    sx := Sgn(dx);
    sy := Sgn(dy);

    pyInc := sy * fPitch;

    x := x1;
    y := y1;
    pptr := fLinesPtr[y];
    Inc(pptr, x);
    if (ax > ay) then begin             (* x dominant *)
      d := ay - (ax shr 1);
      if dx < 0 then begin
        while (x >= x2) do begin
          FillOneSub(pptr, fPenColor24);
          if (d >= 0) then begin
            Inc(PByte(pptr), pyInc);
            Dec(d, ax);
          end;
          Dec(x);
          Dec(pptr);
          Inc(d, ay);
        end
      end
      else begin
        while (x <= x2) do begin
          FillOneSub(pptr, fPenColor24);
          if (d >= 0) then begin
            Inc(PByte(pptr), pyInc);
            Dec(d, ax);
          end;
          Inc(x);
          Inc(pptr);
          Inc(d, ay);
        end
      end
    end
    else begin                          (* y dominant *)
      d := ax - (ay shr 1);
      if dy < 0 then begin
        while (y >= y2) do begin
          FillOneSub(pptr, fPenColor24);
          if (d >= 0) then begin
            Inc(pptr, sx);
            Dec(d, ay);
          end;
          Dec(y);
          Inc(PByte(pptr), pyInc);
          Inc(d, ax);
        end
      end
      else begin
        while (y <= y2) do begin
          FillOneSub(pptr, fPenColor24);
          if (d >= 0) then begin
            Inc(pptr, sx);
            Dec(d, ay);
          end;
          Inc(y);
          Inc(PByte(pptr), pyInc);
          Inc(d, ax);
        end
      end
    end
  end;
  fBitmap.Canvas.PenPos := Point(X2, Y2);
end;

procedure TFastDraw24.PolyLine(const Points: array of TPoint);
var
  i                 : integer;
begin
  fPenColor24 := Color24(fBitmap.Canvas.Pen.Color);
  fBrushColor24 := Color24(fBitmap.Canvas.Brush.Color);

  for i := 0 to Pred(High(Points)) do begin
    LineP(Points[i], Points[i + 1]);
  end;
  fBitmap.Canvas.PenPos := Points[High(Points)];
end;

procedure TFastDraw24.PolyLineClosed(const Points: array of TPoint);
var
  i                 : integer;
begin
  fPenColor24 := Color24(fBitmap.Canvas.Pen.Color);
  fBrushColor24 := Color24(fBitmap.Canvas.Brush.Color);

  for i := 0 to Pred(High(Points)) do begin
    LineP(Points[i], Points[i + 1]);
  end;
  LineP(Points[High(Points)], Points[0]);
  fBitmap.Canvas.PenPos := Points[0];
end;

procedure TFastDraw24.Polygon(const Points: array of TPoint);
var
  a, b, t           : integer;
  k, y0, y1, y, i, j, xl, xr: integer;
  Ind               : array of integer; // list of vertex indices, sorted by pt[ind[j]].y
  e                 : TEdge;
  ep                : PEdge;
begin

  fPenColor24 := Color24(fBitmap.Canvas.Pen.Color);
  fBrushColor24 := Color24(fBitmap.Canvas.Brush.Color);

  if fBitmap.Canvas.Brush.Style = bsDiagCross then begin
    FillSub := Fill24AlfaBlend;
    FillOneSub := FillOne24AlfaBlend;
  end
  else begin
    FillSub := Fill24;
    FillOneSub := FillOne24;
  end;

  fnPts := Succ(High(Points));

  if (fnPts > 2) and (fBitmap.Canvas.Brush.Style <> bsClear) then begin

    SetLength(Ind, fnPts);
    SetLength(fActiveEdge, fnPts);

    // create y-sorted array of indices ind[k] into vertex list
    for k := 0 to Pred(fnPts) do
      Ind[k] := k;

    for a := 0 to Pred(Pred(fnPts)) do begin
      for b := Succ(a) to Pred(fnPts) do begin
        if Points[Ind[b]].y <= Points[Ind[a]].y then begin
          t := Ind[b];
          Ind[b] := Ind[a];
          Ind[a] := t;
        end
      end
    end;

    fnAct := 0;                         // start with empty ActiveEdge list
    y0 := Max(fClipRect.Top, Points[ind[0]].y); // ymin of polygon
    y1 := Min(fClipRect.Bottom, Points[ind[fnPts - 1]].y); // ymax of polygon

    y := y0;
    k := 0;                             // ind[k] is next vertex to process
    while y <= y1 do begin              // step through scanlines
      // check vertices between previous scanline and current one, if any
      while (k < fnPts) and (Points[ind[k]].y <= y) do begin
        i := Ind[k];
        // insert or delete edges before and after vertex i (i-1 to i,
        // and i to i+1) from ActiveEdge list if they cross scanline y
        if (i > 0) then
          j := i - 1
        else
          j := fnPts - 1;               // vertex previous to i

        if (Points[j].y < y) then       // old edge, remove from ActiveEdge list
          CDelete(j)
        else if (Points[j].y > y) then  // new edge, add to ActiveEdge list
          CInsert(Points, j, y);

        if (i < (fnPts - 1)) then
          j := i + 1
        else
          j := 0;                       // vertex next after i

        if (Points[j].y < y) then       // old edge, remove from ActiveEdge list
          CDelete(i)
        else if (Points[j].y > y) then  // new edge, add to ActiveEdge list
          CInsert(Points, i, y);

        Inc(k);
      end;

      // sort ActiveEdge edge list by ActiveEdge[j].x
      for a := 0 to Pred(Pred(fnAct)) do begin
        for b := Succ(a) to Pred(fnAct) do begin
          if fActiveEdge[b].x <= fActiveEdge[a].x then begin
            e := fActiveEdge[b];
            fActiveEdge[b] := fActiveEdge[a];
            fActiveEdge[a] := e;
          end
        end
      end;
      // draw horizontal segments for scanline y
      j := 0;
      while (j < fnAct) do begin        // draw horizontal segments

        // span 'tween j & j+1 is inside, span tween j+1 & j+2 is outside
        xl := fActiveEdge[j].x;         // left end of span
        if (xl < fClipRect.Left) then xl := fClipRect.Left;
        xr := fActiveEdge[j + 1].x;     // right end of span
        if (xr > fClipRect.Right) then xr := fClipRect.Right;
        if (xl <= xr) then begin
          // draw pixels in span
          FillSub(@PRGBTripleArray(fLinesPtr[y])[xl], xr - xl + 1, fBrushColor24);
        end;

        ep := @fActiveEdge[j];
        ep.Acc := ep.Acc + Abs(ep.Inc);
        while ep.Acc >= ep.Tst do begin
          Inc(ep.x, ep.Step);
          ep.Acc := ep.Acc - ep.Tst;
        end;
        Inc(j);

        ep := @fActiveEdge[j];
        Inc(ep.Acc, ep.Inc);
        while ep.Acc >= ep.Tst do begin
          Inc(ep.x, ep.Step);
          ep.Acc := ep.Acc - ep.Tst;
        end;
        Inc(j);

      end;
      Inc(y);
    end
  end
  else begin
    fBrushColor24 := $FFFFFFFF;
  end;
  PolyLineClosed(Points);
  fBitmap.Canvas.PenPos := Points[High(Points)];
end;

constructor TFastDraw24.Create(aBitmap: TBitmap);
begin
  inherited Create(aBitmap);
end;

destructor TFastDraw24.Destroy;
begin
  inherited Destroy;
end;

{ TFastDraw15 }

(*
0 00000 00000 00000 5,5,5 $7BDE 15bit
00000 000000 00000  5,6,5 $F7DE 16bit
*)

procedure FillOne15AlfaBlend(X: Pword; Value: word);
begin
  X^ := ((Value and $7BDE) + (X^ and $7BDE)) shr 1;
end;

procedure Fill15AlfaBlend(X: Pword; Count: Integer; Value: word);
begin
  while Count > 0 do begin
    X^ := ((Value and $7BDE) + (X^ and $7BDE)) shr 1;
    Inc(X);
    Dec(Count);
  end
end;

procedure FillOne16AlfaBlend(X: Pword; Value: word);
begin
  X^ := ((Value and $F7DE) + (X^ and $F7DE)) shr 1;
end;

procedure Fill16AlfaBlend(X: Pword; Count: Integer; Value: word);
begin
  while Count > 0 do begin
    X^ := ((Value and $F7DE) + (X^ and $F7DE)) shr 1;
    Inc(X);
    Dec(Count);
  end
end;

procedure FillOne15(X: Pword; Value: word);
begin
  X^ := Value;
end;

procedure Fill15(X: Pword; Count: Integer; Value: word);
begin
  while Count > 0 do begin
    X^ := Value;
    Inc(X);
    Dec(Count);
  end
end;

function Color15(c: TColor): integer;
var
  cl                : longword;
  //  r, g, b: integer;
begin
  cl := ColorToRGB(c);

  Result := ((cl and $00F80000) shr 19) +
    ((cl and $0000F800) shr 6) +
    ((cl and $000000F8) shl 7);

  (*
    b := ((cl and $00FF0000) shr 16) shr 3;
    g := ((cl and $0000FF00) shr  8) shr 3;
    r := ((cl and $000000FF) shr  0) shr 3;
    Result := b + (g shl 5) + (r shl 10);
  *)
end;

function Color16(c: TColor): integer;
var
  cl                : longword;
  //  r, g, b: integer;
begin
  cl := ColorToRGB(c);
  Result := ((cl and $00F80000) shr 19) +
    ((cl and $0000FC00) shr 5) +
    ((cl and $000000F8) shl 8);

  (*
b := ((cl and $00FF0000) shr 16) shr 3;
g := ((cl and $0000FF00) shr  8) shr 2;
r := ((cl and $000000FF) shr  0) shr 3;
Result := b + (g shl 5) + (r shl 11);
*)
end;

//-------------------------------------------------------------------------

procedure TFastDraw15.RecreateScanLines;
var
  i                 : integer;
begin
  SetLength(fLinesPtr, fBitmap.Height);
  for i := 0 to Pred(fBitmap.Height) do
    fLinesPtr[i] := fBitmap.Scanline[i];
  fPitch := (integer(fLinesPtr[1]) - integer(fLinesPtr[0])) div SizeOf(word);
  fClipRect := Rect(0, 0, fBitmap.Width, fBitmap.Height);
  Inc(fClipRect.Top);
  Inc(fClipRect.Left);
  Dec(fClipRect.Right);
  Dec(fClipRect.Bottom, 2);
end;

procedure TFastDraw15.LineP(p1, p2: TPoint);
var
  d, x, y, ax, ay, sx, sy, dx, dy: integer;
  pptr              : PWord;
  pyInc             : integer;
  Code1, Code2      : integer;

begin

  if fIs16 then
    fPenColor15 := Color16(fBitmap.Canvas.Pen.Color)
  else
    fPenColor15 := Color15(fBitmap.Canvas.Pen.Color);

  dx := p2.x - p1.x;
  dy := p2.y - p1.y;

  //  if ClipLine(p1, p2, fClipRect, dx, dy) then begin
  Code1 := 0;
  Code2 := 0;

  if (p1.x < fClipRect.Left) then begin
    if dx <> 0 then begin
      p1.y := Floor(p1.y + (dy / dx) * (fClipRect.Left - p1.x));
      p1.x := fClipRect.Left;
    end;
    Code1 := 1;
  end
  else if (p1.x > fClipRect.Right) then begin
    if dx <> 0 then begin
      p1.y := Floor(p1.y + (dy / dx) * (fClipRect.Right - p1.x));
      p1.x := fClipRect.Right;
    end;
    Code1 := 2;
  end;

  if (p1.y < fClipRect.Top) then begin
    if dy <> 0 then begin
      p1.x := Floor(p1.x + (dx / dy) * (fClipRect.Top - p1.y));
      p1.y := fClipRect.Top;
    end;
    Code1 := Code1 or 4;
  end
  else if (p1.y > fClipRect.Bottom) then begin
    if dy <> 0 then begin
      p1.x := Floor(p1.x + (dx / dy) * (fClipRect.Bottom - p1.y));
      p1.y := fClipRect.Bottom;
    end;
    Code1 := Code1 or 8;
  end;

  if (p2.x < fClipRect.Left) then begin
    if dx <> 0 then begin
      p2.y := Floor(p2.y + (dy / dx) * (fClipRect.Left - p2.x));
      p2.x := fClipRect.Left;
    end;
    Code2 := 1;
  end
  else if (p2.x > fClipRect.Right) then begin
    if dx <> 0 then begin
      p2.y := Floor(p2.y + (dy / dx) * (fClipRect.Right - p2.x));
      p2.x := fClipRect.Right;
    end;
    Code2 := 2;
  end;

  if (p2.y < fClipRect.Top) then begin
    if dy <> 0 then begin
      p2.x := Floor(p2.x + (dx / dy) * (fClipRect.Top - p2.y));
      p2.y := fClipRect.Top;
    end;
    Code2 := Code2 or 4;
  end
  else if (p2.y > fClipRect.Bottom) then begin
    if dy <> 0 then begin
      p2.x := Floor(p2.x + (dx / dy) * (fClipRect.Bottom - p2.y));
      p2.y := fClipRect.Bottom;
    end;
    Code2 := Code2 or 8;
  end;

  if ((Code1 and Code2) = 0) and
    (p1.x >= fClipRect.Left) and
    (p1.x <= fClipRect.Right) then begin

    // ---------------------------------------------------
    ax := Abs(dx) shl 1;
    ay := Abs(dy) shl 1;

    sx := Sgn(dx);
    sy := Sgn(dy);

    pyInc := sy * fPitch;

    x := p1.x;
    y := p1.y;
    pptr := fLinesPtr[y];
    Inc(pptr, x);
    if (ax > ay) then begin             (* x dominant *)
      d := ay - (ax shr 1);
      if dx < 0 then begin
        while (x >= p2.x) do begin
          //          pptr^ := fPenColor15;
          FillOneSub(pptr, fPenColor15);
          //          if (x = p2.x) then
          //            Exit;
          if (d >= 0) then begin
            //            Inc(y, sy);
            Inc(pptr, pyInc);
            Dec(d, ax);
          end;
          Dec(x);
          Dec(pptr);
          Inc(d, ay);
        end
      end
      else begin
        while (x <= p2.x) do begin
          //          pptr^ := fPenColor15;
          FillOneSub(pptr, fPenColor15);
          //          if (x = p2.x) then
          //            Exit;
          if (d >= 0) then begin
            //            Inc(y, sy);
            Inc(pptr, pyInc);
            Dec(d, ax);
          end;
          Inc(x);
          Inc(pptr);
          Inc(d, ay);
        end
      end
    end
    else begin                          (* y dominant *)
      d := ax - (ay shr 1);
      if dy < 0 then begin
        while (y >= p2.y) do begin
          //          pptr^ := fPenColor15;
          FillOneSub(pptr, fPenColor15);
          //          if (y = p2.y) then
          //            Exit;
          if (d >= 0) then begin
            //            Inc(x, sx);
            Inc(pptr, sx);
            Dec(d, ay);
          end;
          Dec(y);
          Inc(pptr, pyInc);
          Inc(d, ax);
        end
      end
      else begin
        while (y <= p2.y) do begin
          //          pptr^ := fPenColor15;
          FillOneSub(pptr, fPenColor15);
          //          if (y = p2.y) then
          //            Exit;
          if (d >= 0) then begin
            //            Inc(x, sx);
            Inc(pptr, sx);
            Dec(d, ay);
          end;
          Inc(y);
          Inc(pptr, pyInc);
          Inc(d, ax);
        end
      end
    end
  end
end;

procedure TFastDraw15.LineC(x1, y1, x2, y2: integer);
var
  d, x, y, ax, ay, sx, sy, dx, dy: integer;
  pptr              : Pword;
  pyInc             : integer;
  Code1, Code2      : integer;

begin

  if fIs16 then
    fPenColor15 := Color16(fBitmap.Canvas.Pen.Color)
  else
    fPenColor15 := Color15(fBitmap.Canvas.Pen.Color);

  dx := x2 - x1;
  dy := y2 - y1;

  //  if ClipLine(p1, p2, fClipRect, dx, dy) then begin
  Code1 := 0;
  Code2 := 0;

  if (x1 < fClipRect.Left) then begin
    if dx <> 0 then begin
      y1 := Floor(y1 + (dy / dx) * (fClipRect.Left - x1));
      x1 := fClipRect.Left;
    end;
    Code1 := 1;
  end
  else if (x1 > fClipRect.Right) then begin
    if dx <> 0 then begin
      y1 := Floor(y1 + (dy / dx) * (fClipRect.Right - x1));
      x1 := fClipRect.Right;
    end;
    Code1 := 2;
  end;

  if (y1 < fClipRect.Top) then begin
    if dy <> 0 then begin
      x1 := Floor(x1 + (dx / dy) * (fClipRect.Top - y1));
      y1 := fClipRect.Top;
    end;
    Code1 := Code1 or 4;
  end
  else if (y1 > fClipRect.Bottom) then begin
    if dy <> 0 then begin
      x1 := Floor(x1 + (dx / dy) * (fClipRect.Bottom - y1));
      y1 := fClipRect.Bottom;
    end;
    Code1 := Code1 or 8;
  end;

  if (x2 < fClipRect.Left) then begin
    if dx <> 0 then begin
      y2 := Floor(y2 + (dy / dx) * (fClipRect.Left - x2));
      x2 := fClipRect.Left;
    end;
    Code2 := 1;
  end
  else if (x2 > fClipRect.Right) then begin
    if dx <> 0 then begin
      y2 := Floor(y2 + (dy / dx) * (fClipRect.Right - x2));
      x2 := fClipRect.Right;
    end;
    Code2 := 2;
  end;

  if (y2 < fClipRect.Top) then begin
    if dy <> 0 then begin
      x2 := Floor(x2 + (dx / dy) * (fClipRect.Top - y2));
      y2 := fClipRect.Top;
    end;
    Code2 := Code2 or 4;
  end
  else if (y2 > fClipRect.Bottom) then begin
    if dy <> 0 then begin
      x2 := Floor(x2 + (dx / dy) * (fClipRect.Bottom - y2));
      y2 := fClipRect.Bottom;
    end;
    Code2 := Code2 or 8;
  end;

  if ((Code1 and Code2) = 0) and
    (x1 >= fClipRect.Left) and
    (x1 <= fClipRect.Right) then begin

    // ---------------------------------------------------
    ax := Abs(dx) shl 1;
    ay := Abs(dy) shl 1;

    sx := Sgn(dx);
    sy := Sgn(dy);

    pyInc := sy * fPitch;

    x := x1;
    y := y1;
    pptr := fLinesPtr[y];
    Inc(pptr, x);
    if (ax > ay) then begin             (* x dominant *)
      d := ay - (ax shr 1);
      if dx < 0 then begin
        while (x >= x2) do begin
          //        pptr^ := fPenColor15;
          FillOneSub(pptr, fPenColor15);
          //          if (x = x2) then
          //            Exit;
          if (d >= 0) then begin
            //            Inc(y, sy);
            Inc(pptr, pyInc);
            Dec(d, ax);
          end;
          Dec(x);
          Dec(pptr);
          Inc(d, ay);
        end
      end
      else begin
        while (x <= x2) do begin
          //        pptr^ := fPenColor15;
          FillOneSub(pptr, fPenColor15);
          //          if (x = x2) then
          //            Exit;
          if (d >= 0) then begin
            //            Inc(y, sy);
            Inc(pptr, pyInc);
            Dec(d, ax);
          end;
          Inc(x);
          Inc(pptr);
          Inc(d, ay);
        end
      end
    end
    else begin                          (* y dominant *)
      d := ax - (ay shr 1);
      if dy < 0 then begin
        while (y >= y2) do begin
          //        pptr^ := fPenColor15;
          FillOneSub(pptr, fPenColor15);
          //          if (y = y2) then
          //            Exit;
          if (d >= 0) then begin
            //            Inc(x, sx);
            Inc(pptr, sx);
            Dec(d, ay);
          end;
          Dec(y);
          Inc(pptr, pyInc);
          Inc(d, ax);
        end
      end
      else begin
        while (y <= y2) do begin
          //        pptr^ := fPenColor15;
          FillOneSub(pptr, fPenColor15);
          //          if (y = y2) then
          //            Exit;
          if (d >= 0) then begin
            //            Inc(x, sx);
            Inc(pptr, sx);
            Dec(d, ay);
          end;
          Inc(y);
          Inc(pptr, pyInc);
          Inc(d, ax);
        end
      end
    end
  end;
  fBitmap.Canvas.PenPos := Point(X2, Y2);
end;

procedure TFastDraw15.PolyLine(const Points: array of TPoint);
var
  i                 : integer;
begin
  if fIs16 then begin
    fPenColor15 := Color16(fBitmap.Canvas.Pen.Color);
    fBrushColor15 := Color16(fBitmap.Canvas.Brush.Color);
  end
  else begin
    fPenColor15 := Color15(fBitmap.Canvas.Pen.Color);
    fBrushColor15 := Color15(fBitmap.Canvas.Brush.Color);
  end;

  for i := 0 to Pred(High(Points)) do begin
    LineP(Points[i], Points[i + 1]);
  end;
  fBitmap.Canvas.PenPos := Points[High(Points)];
end;

procedure TFastDraw15.PolyLineClosed(const Points: array of TPoint);
var
  i                 : integer;
begin
  if fIs16 then begin
    fPenColor15 := Color16(fBitmap.Canvas.Pen.Color);
    fBrushColor15 := Color16(fBitmap.Canvas.Brush.Color);
  end
  else begin
    fPenColor15 := Color15(fBitmap.Canvas.Pen.Color);
    fBrushColor15 := Color15(fBitmap.Canvas.Brush.Color);
  end;

  for i := 0 to Pred(High(Points)) do begin
    LineP(Points[i], Points[i + 1]);
  end;
  LineP(Points[High(Points)], Points[0]);
  fBitmap.Canvas.PenPos := Points[0];
end;

procedure TFastDraw15.Polygon(const Points: array of TPoint);
var
  a, b, t           : integer;
  k, y0, y1, y, i, j, xl, xr: integer;
  Ind               : array of integer; // list of vertex indices, sorted by pt[ind[j]].y
  e                 : TEdge;
  ep                : PEdge;
begin

  if fIs16 then begin
    fPenColor15 := Color16(fBitmap.Canvas.Pen.Color);
    fBrushColor15 := Color16(fBitmap.Canvas.Brush.Color);
  end
  else begin
    fPenColor15 := Color15(fBitmap.Canvas.Pen.Color);
    fBrushColor15 := Color15(fBitmap.Canvas.Brush.Color);
  end;

  if fBitmap.Canvas.Brush.Style = bsDiagCross then begin
    if fIs16 then begin
      FillSub := Fill16AlfaBlend;
      FillOneSub := FillOne16AlfaBlend;
    end
    else begin
      FillSub := Fill15AlfaBlend;
      FillOneSub := FillOne15AlfaBlend;
    end
  end
  else begin
    FillSub := Fill15;
    FillOneSub := FillOne15;
  end;

  fnPts := Succ(High(Points));

  if (fnPts > 2) and (fBitmap.Canvas.Brush.Style <> bsClear) then begin

    SetLength(Ind, fnPts);
    SetLength(fActiveEdge, fnPts);

    // create y-sorted array of indices ind[k] into vertex list
    for k := 0 to Pred(fnPts) do
      Ind[k] := k;

    for a := 0 to Pred(Pred(fnPts)) do begin
      for b := Succ(a) to Pred(fnPts) do begin
        if Points[Ind[b]].y <= Points[Ind[a]].y then begin
          t := Ind[b];
          Ind[b] := Ind[a];
          Ind[a] := t;
        end
      end
    end;

    fnAct := 0;                         // start with empty ActiveEdge list
    y0 := Max(fClipRect.Top, Points[ind[0]].y); // ymin of polygon
    y1 := Min(fClipRect.Bottom, Points[ind[fnPts - 1]].y); // ymax of polygon

    y := y0;
    k := 0;                             // ind[k] is next vertex to process
    while y <= y1 do begin              // step through scanlines
      // check vertices between previous scanline and current one, if any
      while (k < fnPts) and (Points[ind[k]].y <= y) do begin
        i := Ind[k];
        // insert or delete edges before and after vertex i (i-1 to i,
        // and i to i+1) from ActiveEdge list if they cross scanline y
        if (i > 0) then
          j := i - 1
        else
          j := fnPts - 1;               // vertex previous to i

        if (Points[j].y < y) then       // old edge, remove from ActiveEdge list
          CDelete(j)
        else if (Points[j].y > y) then  // new edge, add to ActiveEdge list
          CInsert(Points, j, y);

        if (i < (fnPts - 1)) then
          j := i + 1
        else
          j := 0;                       // vertex next after i

        if (Points[j].y < y) then       // old edge, remove from ActiveEdge list
          CDelete(i)
        else if (Points[j].y > y) then  // new edge, add to ActiveEdge list
          CInsert(Points, i, y);

        Inc(k);
      end;

      // sort ActiveEdge edge list by ActiveEdge[j].x
      for a := 0 to Pred(Pred(fnAct)) do begin
        for b := Succ(a) to Pred(fnAct) do begin
          if fActiveEdge[b].x <= fActiveEdge[a].x then begin
            e := fActiveEdge[b];
            fActiveEdge[b] := fActiveEdge[a];
            fActiveEdge[a] := e;
          end
        end
      end;
      // draw horizontal segments for scanline y
      j := 0;
      while (j < fnAct) do begin        // draw horizontal segments

        // span 'tween j & j+1 is inside, span tween j+1 & j+2 is outside
        xl := fActiveEdge[j].x;         // left end of span
        if (xl < fClipRect.Left) then xl := fClipRect.Left;
        xr := fActiveEdge[j + 1].x;     // right end of span
        if (xr > fClipRect.Right) then xr := fClipRect.Right;
        if (xl <= xr) then begin
          // draw pixels in span
          (**)
          FillSub(PWord(longword(fLinesPtr[y]) + xl * SizeOf(word)), xr - xl + 1, fBrushColor15);
          (**)
          (*
                  pptr := fLinesPtr[y];
                  Inc(pptr, xl);
                  while xl <= xr do begin
                    pptr^ := cb;
                    Inc(pptr);
                    Inc(xl);
                  end;
          *)
        end;

        ep := @fActiveEdge[j];
        ep.Acc := ep.Acc + Abs(ep.Inc);
        while ep.Acc >= ep.Tst do begin
          Inc(ep.x, ep.Step);
          ep.Acc := ep.Acc - ep.Tst;
        end;
        Inc(j);

        ep := @fActiveEdge[j];
        Inc(ep.Acc, ep.Inc);
        while ep.Acc >= ep.Tst do begin
          Inc(ep.x, ep.Step);
          ep.Acc := ep.Acc - ep.Tst;
        end;
        Inc(j);

      end;
      Inc(y);
    end
  end
  else begin
    fBrushColor15 := $FFFF;
  end;
  PolyLineClosed(Points);
  fBitmap.Canvas.PenPos := Points[High(Points)];
end;

constructor TFastDraw15.Create(aBitmap: TBitmap);
begin
  inherited Create(aBitmap);
end;

destructor TFastDraw15.Destroy;
begin
  inherited Destroy;
end;

{ TDrawObject }

procedure TDrawObject.SetupCanvas(aCanvas: TCanvas);
begin
  fCanvas := aCanvas;
  CopyRect := aCanvas.CopyRect;
  Draw := aCanvas.Draw;
  Ellipse := aCanvas.Ellipse;
  FillRect := aCanvas.FillRect;
  FrameRect := aCanvas.FrameRect;
  Line := LineC;
  LineTo := aCanvas.LineTo;
  MoveTo := aCanvas.MoveTo;
  Polygon := aCanvas.Polygon;
  Polyline := aCanvas.PolyLine;
  Rectangle := aCanvas.Rectangle;
  Refresh := aCanvas.Refresh;
  StretchDraw := aCanvas.StretchDraw;
  TextHeight := aCanvas.TextHeight;
  TextOut := aCanvas.TextOut;
  TextRect := aCanvas.TextRect;
  TextWidth := aCanvas.TextWidth;
end;

constructor TDrawObject.Create(aCanvas: TCanvas);
begin
  SetupCanvas(aCanvas);
  fFastDraw := nil;
end;

constructor TDrawObject.Create(aBitmap: TBitmap);
begin
  SetupCanvas(aBitmap.Canvas);
  if (aBitmap.PixelFormat in [pf15Bit, pf16Bit, pf24Bit, pf32Bit]) then begin
    case aBitmap.PixelFormat of
      pf32Bit: begin
          fFastDraw := TFastDraw32.Create(aBitmap);
        end;
      pf24Bit: begin
          fFastDraw := TFastDraw24.Create(aBitmap);
        end;
      pf15Bit: begin
          fFastDraw := TFastDraw15.Create(aBitmap);
          fFastDraw.fIs16 := False;
        end;
      pf16Bit: begin
          fFastDraw := TFastDraw15.Create(aBitmap);
          fFastDraw.fIs16 := True;
        end;
    end;
    //    FillRect := fFastDraw.FillRect;
    Line := fFastDraw.LineC;
    LineTo := fFastDraw.LineTo;
    MoveTo := fFastDraw.MoveTo;
    Polygon := fFastDraw.Polygon;
    Polyline := fFastDraw.PolyLine;
    Rectangle := fFastDraw.Rectangle;
  end
end;

destructor TDrawObject.Destroy;
begin
  fCanvas := nil;
  inherited;
  fFastDraw.Free;
end;

procedure TDrawObject.BeginUpdate;
begin
  if Assigned(fFastDraw) then
    fFastDraw.RecreateScanLines;
end;

procedure TDrawObject.EndUpdate;
begin
  // nothing
end;

procedure TDrawObject.LineC(X1, Y1, X2, Y2: integer);
begin
  MoveTo(X1, Y1);
  LineTo(X2, Y2);
end;

end.

